#-----------------------------
# olegen.rb
# $Date: 2002/03/23 06:54:11 $
# $Revision: 1.1 $
#-----------------------------

require 'win32ole'

class WIN32COMGen
  def initialize(typelib)
    @typelib = typelib
    @reciever = ""
  end
  attr_reader :typelib

  def ole_classes(typelib)
    begin
      @ole = WIN32OLE.new(typelib)
      [@ole.ole_obj_help]
    rescue
      WIN32OLE_TYPE.ole_classes(typelib)
    end
  end

  def generate_args(method)
    args = []
    if method.size_opt_params >= 0
      size_required_params = method.size_params - method.size_opt_params
    else
      size_required_params = method.size_params - 1
    end
    size_required_params.times do |i|
      args.push "arg#{i}"
    end
    if method.size_opt_params >= 0
      method.size_opt_params.times do |i|
        args.push "arg#{i + size_required_params}=nil"
      end
    else
      args.push "*arg"
    end
    args.join(", ")
  end

  def generate_argtypes(method)
    method.params.collect{|param|
      ts = ''
      param.ole_type_detail.each do |t|
	case t
	when 'CARRAY', 'VOID', 'INT', 'UINT', 'RESULT', 'DECIMAL', 'I8', 'UI8', 'INT', 'UINT'
#	  raise "Sorry type\"" + t + "\" not supported"
          ts << "\"??? NOT SUPPORTED TYPE:`#{t}'\""
	when 'USERDEFINED', 'Unknown Type 9'
	  ts << 'VT_DISPATCH'
          break;
	when 'SAFEARRAY'
	  ts << 'VT_ARRAY|'
	when 'PTR'
	  ts << 'VT_BYREF|'
	else
	  if String === t
	    ts << 'VT_' + t
	  end
	end
      end
      if ts.empty?
        ts = 'VT_VARIANT'
      elsif ts[-1] == ?|
	ts += 'VT_VARIANT'
      end
      ts
    }.join(", ")
  end

  def generate_method_body(method)
    "    ret = #{@reciever}_invoke(#{method.dispid}, [" +
    generate_args(method).gsub("=nil", "") +
    "], [" +
    generate_argtypes(method) +
    "])\n" +
    "    @lastargs = WIN32OLE::ARGV\n" +
    "    ret"
  end

  def generate_method(method, io = STDOUT)
    io.puts  "\n  # #{method.helpstring}"
    io.print "  def #{method.name}("
    io.print generate_args(method)
    io.puts ")"
    io.puts generate_method_body(method)
    io.puts "  end"
  end

  def generate_methods(klass, io = STDOUT)
    klass.ole_methods.select {|method|
      method.invoke_kind == "FUNC" && method.visible?
    }.each do |method|
      generate_method(method, io)
    end
  end

  def class_name(klass)
    klass_name = klass.name
    if klass.ole_type == "Class" &&
       klass.guid &&
       klass.progid
       klass_name = klass.progid.gsub(/\./, '_')
    end
    if /^[A-Z]/ !~ klass_name
      klass_name = 'OLE' + klass_name
    end
    klass_name
  end

  def define_initialize(klass)
    <<STR

  def initialize(obj = nil)
    @clsid = "#{klass.guid}"
    @progid = "#{klass.progid}"
    if obj.nil?
      @dispatch = WIN32OLE.new @progid
    else
      @dispatch = obj
    end
  end
STR
  end

  def define_include
    "  include WIN32OLE::VARIANT"
  end

  def define_instance_variables
    "  attr_reader :lastargs"
  end

  def define_method_missing
    <<STR

  def method_missing(cmd, *arg)
    @dispatch.method_missing(cmd, *arg)
  end
STR
  end

  def define_class(klass, io = STDOUT)
    io.puts "class #{class_name(klass)}"
    io.puts define_include
    io.puts define_instance_variables
    io.puts "  attr_reader :dispatch"
    io.puts "  attr_reader :clsid"
    io.puts "  attr_reader :progid"
    io.puts define_initialize(klass)
    io.puts define_method_missing
  end

  def define_module(klass, io = STDOUT)
    io.puts "module #{class_name(klass)}"
    io.puts define_include
    io.puts define_instance_variables
  end

  def generate_class(klass, io = STDOUT)
    io.puts "\n# #{klass.helpstring}"
    if klass.ole_type == "Class" &&
       klass.guid &&
       klass.progid
      @reciever = "@dispatch."
      define_class(klass, io)
    else
      @reciever = ""
      define_module(klass, io)
    end
    generate_methods(klass, io)
    io.puts "end"
  end

  def generate(io = STDOUT)
    io.puts "require 'win32ole'"

    ole_classes(typelib).select{|klass|
      klass.visible? &&
      (klass.ole_type == "Class" || 
       klass.ole_type == "Interface" || 
       klass.ole_type == "Dispatch" )
    }.each do |klass|
      generate_class(klass, io)
    end
    begin
      @ole.quit if @ole
    rescue 
    end
  end
end

require 'win32ole'
if __FILE__ == $0
  if ARGV.size == 0
    $stderr.puts "usage: #{$0} Type Library [...]"
    exit 1
  end
  ARGV.each do |typelib|
    comgen = WIN32COMGen.new(typelib)
    comgen.generate
  end
end
